from typing import Any, Self

from pydantic import BaseModel as _BaseModel
from pydantic import model_validator

from ..utils import format_str


class BaseModel(_BaseModel):
    @property
    def fields(self) -> dict[str, Any]:
        """Return all fields of the model as a dictionary."""
        schema = self.model_fields
        field_names = schema.keys()
        field_values = {name: getattr(self, name) for name in field_names}
        return field_values

    @model_validator(mode="after")
    def _format_fields(self) -> Self:
        fields_to_format = {"name", "description", "story", "text"}

        for field_name, field_value in self.fields.items():
            if field_name in fields_to_format and isinstance(field_value, str):
                setattr(self, field_name, format_str(field_value))

        return self



from pydantic import Field, field_validator

from .base import BaseModel

__all__ = ("Changelog", "ChangelogCategory")


class ChangelogCategory(BaseModel):
    category: str
    item_ids: list[int]

    @field_validator("item_ids", mode="before")
    def _intify_ids(cls, v: list[str]) -> list[int]:
        return [int(i) for i in v]


class Changelog(BaseModel):
    id: int
    version: str
    categories: list[ChangelogCategory] = Field(alias="items")
    beta: bool = Field(False)

    @field_validator("categories", mode="before")
    def _convert_categories(cls, v: dict[str, list[int]]) -> list[ChangelogCategory]:
        return [ChangelogCategory(category=k, item_ids=v) for k, v in v.items()]



import re


def format_str(text: str) -> str:
    clean = re.compile(r"<.*?>|\{SPRITE_PRESET#[^\}]+\}")
    return remove_ruby_tags(replace_pronouns(re.sub(clean, "", text).replace("\\n", "\n")))


def find_next_letter(text: str, placeholder: str) -> str:
    """Find the next letter after a placeholder in a string"""
    index = text.find(placeholder)
    index += len(placeholder)
    return text[index]


def replace_placeholders(
    string: str, params: dict[str, list[float | int]] | list[int] | None
) -> str:
    if params is None:
        return string
    if isinstance(params, list):
        for i, value in enumerate(params):
            placeholder = f"#i[{i}]"
            if placeholder in string:
                value_ = value * 100 if find_next_letter(string, placeholder) == "%" else value
                string = string.replace(placeholder, str(value_))
        return string
    for key, values in params.items():
        value = values[0]
        placeholder = f"#{key}[i]"
        if placeholder in string:
            if find_next_letter(string, placeholder) == "%":
                value *= 100
            string = string.replace(placeholder, str(value))
    return string


def replace_pronouns(text: str) -> str:
    female_pronoun_pattern = r"\{F#(.*?)\}"
    male_pronoun_pattern = r"\{M#(.*?)\}"

    female_pronoun_match = re.search(female_pronoun_pattern, text)
    male_pronoun_match = re.search(male_pronoun_pattern, text)

    if female_pronoun_match and male_pronoun_match:
        female_pronoun = female_pronoun_match.group(1)
        male_pronoun = male_pronoun_match.group(1)
        replacement = f"{female_pronoun}/{male_pronoun}"

        text = re.sub(female_pronoun_pattern, replacement, text)
        text = re.sub(male_pronoun_pattern, "", text)
        text = text.replace("#", "")

    return text


def remove_ruby_tags(text: str) -> str:
    # Remove {RUBY_E#} tags
    text = re.sub(r"\{RUBY_E#\}", "", text)
    # Remove {RUBY_B...} tags
    text = re.sub(r"\{RUBY_B[^}]*\}", "", text)
    return text



from typing import Any

from pydantic import Field, field_validator

from .base import BaseModel

__all__ = (
    "LightCone",
    "LightConeAscensionMaterial",
    "LightConeCostItem",
    "LightConeDetail",
    "LightConePathType",
    "LightConeSkill",
    "LightConeUpgrade",
)


class LightConeAscensionMaterial(BaseModel):
    id: int
    rarity: int


class LightConeSkill(BaseModel):
    name: str
    description: str
    params: dict[str, list[int | float]]


class LightConeCostItem(BaseModel):
    id: int
    amount: int


class LightConeUpgrade(BaseModel):
    level: int
    cost_items: list[LightConeCostItem] = Field(alias="costItems")
    max_level: int = Field(alias="maxLevel")
    required_player_level: int = Field(alias="playerLevelRequire")
    required_world_level: int = Field(alias="worldLevelRequire")
    skill_base: dict[str, int | float] = Field(alias="skillBase")
    skill_add: dict[str, int | float] = Field(alias="skillAdd")

    @field_validator("required_player_level", mode="before")
    def _convert_required_player_level(cls, v: int | None) -> int:
        return v or 0

    @field_validator("required_world_level", mode="before")
    def _convert_world_level_require(cls, v: int | None) -> int:
        return v or 0

    @field_validator("cost_items", mode="before")
    def _convert_cost_items(cls, v: dict[str, int] | None) -> list[LightConeCostItem]:
        return (
            [LightConeCostItem(id=int(id_), amount=amount) for id_, amount in v.items()]
            if v
            else []
        )


class LightConePathType(BaseModel):
    id: str
    name: str


class LightConeDetail(BaseModel):
    id: int
    name: str
    beta: bool = Field(False)
    rarity: int = Field(alias="rank")
    type: LightConePathType = Field(alias="types")
    icon: str
    is_sellable: bool = Field(alias="isSellable")
    route: str
    description: str
    upgrades: list[LightConeUpgrade] = Field(alias="upgrade")
    skill: LightConeSkill
    ascension_materials: list[LightConeAscensionMaterial] = Field(alias="ascension")

    @field_validator("type", mode="before")
    def _convert_type(cls, v: dict[str, dict[str, Any]]) -> LightConePathType:
        return LightConePathType(**v["pathType"])

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/equipment/{v}.png"

    @field_validator("upgrades", mode="before")
    def _convert_upgrades(cls, v: list[dict[str, Any]]) -> list[LightConeUpgrade]:
        return [LightConeUpgrade(**upgrade) for upgrade in v]

    @field_validator("ascension_materials", mode="before")
    def _convert_ascension_materials(cls, v: dict[str, int]) -> list[LightConeAscensionMaterial]:
        return [LightConeAscensionMaterial(id=int(id_), rarity=rarity) for id_, rarity in v.items()]

    @property
    def medium_icon(self) -> str:
        return self.icon.replace("equipment", "equipment/medium")

    @property
    def large_icon(self) -> str:
        return self.icon.replace("equipment", "equipment/large")


class LightCone(BaseModel):
    id: int
    name: str
    beta: bool = Field(False)
    rarity: int = Field(alias="rank")
    icon: str
    type: str = Field(alias="types")
    is_sellable: bool = Field(alias="isSellable")
    route: str

    @field_validator("icon", mode="before")
    def _convert_icon(cls, v: str) -> str:
        return f"https://api.yatta.top/hsr/assets/UI/equipment/{v}.png"

    @field_validator("type", mode="before")
    def _convert_type(cls, v: dict[str, str]) -> str:
        return v["pathType"]

    @property
    def medium_icon(self) -> str:
        return self.icon.replace("equipment", "equipment/medium")

    @property
    def large_icon(self) -> str:
        return self.icon.replace("equipment", "equipment/large")
